function FVSWE1Dv3
% File name: FVSWE1Dv3.m
% Author: Clive Mingham
% Date: 18 Oct 2011
% This m-file is associated with the free text book: 'Introductory Finite 
% Volume Methods for PDEs' which may be downloaded from bookboon.com
% when published.
%
% Notes: 
% 1. This prog is version 3 of a sequence of programs which gradually 
% build up to coding a FV solver on a general 2D structured mesh.  
% 
% 2. This code is based on FVSWE1Dv2.m by Clive Mingham.  This is the same
% code except that interface fluxes are obtained by linear interpolation 
% on a Cartesian mesh (which is the same as averaging neighbouring values
% on this mesh).
%
% Description:  Solves the system of 1D PDEs, dU/dt + div(H) = 0,
% using a finite volume scheme on a single row of uniform Cartesian cells 
% in the x direction.  
%
% This equation corresponds to the finite volume form:
%
% doubleintegral(dU/dt dA) + lineintegral(H.n ds) = 0.
%
% In this case the equations are the 1D shallow water equations (SWE) so: 
% Water depth: h=h(t,x) [m].
% Water velocity: v = vx i + 0 j where vx=vx(t,x) [m/s].
% Acceleration due to gravity: g=9.81 [m/s^2].
% Geopotential: phi = gh. 
% Matrix of conserved variables: U = [ phi
%                                     phi vx].
% Flux density: H = [       phi v           i + [0   j
%                    phi vx^2 + 0.5*phi^2 ]      0]
%
% subfunctions: freadmesh, fcellareas, fsidevectors, finitialU, 
%               finterfaceflux, finsertghostcells, fcalcdt, 
%               fsolver, fremoveghostcells, fplotresults, fgetHifromUi.
%
% Verification:  Looks same as version 2.
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%
clc; clf; clear all; % clear everything
runtime=1.2;  % runtime [s]
t=0;          % current time [s]
timelevel=0;  % current time level
[x,y,xcen,ycen]=freadmesh; % gets coordinates of the computational mesh
                           % and cell centres
A=fcellareas(x,y);   % compute and store areas for each cell.  Includes
                     % left and right ghost cells.
S=fsidevectors(x,y); % compute and store cell side vectors in an
                     % (NI+2)x2x2 array which contains left and 
                     %  right ghost cells.
%
U=finitialU(xcen); % Put initial cell centre values of U in a 2xNI array
%
Uinitial=U;  % Store initial profile for plotting if required
U=finsertghostcells(U); % Append ghost cells to each end of the domain. 
                        % This means that indices of computational cell 
                        % start at i=2 and end at i = NI+1.
Unext=zeros(size(U)); % Creates an array for the updated U values 
[d,N]=size(U);        % d is the number of PDEs (=rows of U)
NI=N-2; % number of computational cells in the x direction
%%
disp('time marching starts') 
while(t<runtime)   
 timelevel=timelevel+1 
 U=fbcs(U);   % Insert boundary conditions into ghost cells at each end
              % of the domain. U becomes 2x(NI+2) so each computational 
              % cell is indexed i=2 to NI+1.
 dt=fcalcdt(A,S,U); % Finds stable time step before each iteration.
 t=t+dt;  % updates time  
 if (t>runtime) % ensures program stops at exactly t=runtime [s]
     dt=dt-(t-runtime);
     t=runtime;
 end
 t   % Outputs current time to the Command Window
 dt
 Unext=fsolver(U,A,dt,S,x); % implements a time step of the solver
%
 U(:,2:NI+1)=Unext(:,2:NI+1);  % update computational U values prior to
                               % the next time iteration
end  % of while loop
disp('1D finite volume scheme ended at time t')
t
U=fremoveghostcells(U);
fplotresults(xcen,U) % plot results
disp('finished main program, see figures')
end % FVSWE1Dv3
%%%%%%%%%%%%%%%%%%%%%%%%%%% subfunctions  %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
function [x,y,xcen,ycen]=freadmesh
% Verification:  verified.
% The mesh has NI uniform quadrilateral cells in the x direction and 
% NJ=1 cell in the j direction (for a 1D problem).
% The x and y coordinates of the lower left hand corner of cell (i,j) are
% held in arrays x and y respectively which are both (NI+1)x(NJ+1).  In
% this way the 4 vertices of cell (i,j) are (x(i),y(j)), (x(i+1),y(j)),
% (x(i+1),y(j+1)) and (x(i),y(j+1)).  
% The mesh starts at the origin so x(1,1)=0=y(1,1).
% 
NI=100;  % number of cells in the i direction
NJ=1;    % number of cells in the j direction (=1 for a 1D problem)
x=zeros(NI+1,NJ+1); % allocate correct sized array for cell vertices
y=zeros(NI+1,NJ+1); % allocate correct sized array for cell vertices
xcen=zeros(NI,NJ); % allocate correct sized array for cell centres
ycen=zeros(NI,NJ); % allocate correct sized array for cell centres
%
dx=1.0; % cell length in x direction
dy=2.0; % cell length in y direction (arbitrary)
%
x(1,1)=0; % first row starting point for x
y(1,1)=0; % first row starting point for x
x(1,2)=0; % second row starting point for x
y(1,2)=dy;  % second row starting point for y
% create uniform mesh
for i=2:NI+1
      x(i,1)=x(1,1)+(i-1)*dx;
      y(i,1)=y(1,1);
      x(i,2)=x(1,2)+(i-1)*dx;
      y(i,2)=y(1,2);
end % of i loop
% find cell centres by averaging coordinates of vertices 
% (this works for a general structured mesh)
for i=1:NI
     xcen(i,1)=(x(i,1)+x(i+1,1)+x(i+1,2)+x(i,2))/4;
     ycen(i,1)=(y(i,1)+y(i+1,1)+y(i+1,2)+y(i,2))/4;
end % of i loop
end % freadmesh
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
function A=fcellareas(x,y)
% General way to find quadrilateral cell areas.
% Defines A to include left and right ghost cells.
 [m,n]=size(x);
 NI=m-1; % number of computational cells in the x direction
 NJ=n-1; % number of computational cells in the y direction
 A=zeros(NI+2);  % allocate correct size array including ghost cells
 % inclusion of a left ghost cell means that indices of computational cells
 % go from i=2 to i=NI+1
 for i=1:NI
     for j=1:NJ % j goes from 1 to 1
         % write the diagonal vectors as 3D vectors for the cross product
         d1=[x(i+1,j+1)-x(i,j),y(i+1,j+1)-y(i,j),0]; 
         d2=[x(i,j+1)-x(i+1,j),y(i,j+1)-y(i+1,j),0];
         d=cross(d1,d2);   
         A(i+1)=0.5*norm(d); % note shift of i index by 1 in A
     end % of j loop
 end % of i loop
end % of fcellareas
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%% 
function S=fsidevectors(x,y)
% Verification: 
%
% Finds the left and right side vectors for each cell and stores in array
% S.  The first component of S holds the cell index.  The second component
% of S indicates the right (entry = 1) or left (entry = 2) side vector.
% The third component of S indicates the i (entry = 1) or j (entry = 2) 
% component of the side vector.
%
% If a i + b j is the vector representing the cell side then the 
% corresponding side vector is b i - a j.
% Notes:
% 1. S contains left and right ghost cells.
%
% 2. vectors are 1 by 2 arrays so care must be taken because side
%    vectors are stored in higher dimensional arrays.
%
[m,n]=size(x);
NI=m-1; % number of computational cells in the x direction
NJ=n-1; % number of computational cells in the y direction
S=zeros(NI+2,2,2); % includes ghost cells
% ghost cells shift the i index by 1 so computational indices
% go from 2 to NI+1
for i=1:NI
    for j=1:NJ       
        rightcellside=[x(i+1,j+1)-x(i+1,j),y(i+1,j+1)-y(i+1,j)];
        S(i+1,1,1)=rightcellside(2); % note shift of i index
        S(i+1,1,2)=-rightcellside(1);       
        %
        leftcellside=[x(i,j)-x(i,j+1),y(i,j)-y(i,j+1)];
        S(i+1,2,1)=leftcellside(2);
        S(i+1,2,2)=-leftcellside(1);    
    end % of j loop
end % of i loop
end % fsidevectors
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
function U=finitialU(xcen)
% This sets up a dam break problem.
% Verification:  correct.
% Inserts initial U values in computational cells (U is 2xNI)
NI=length(xcen); % number of computational cells in x direction
%
d=2;  % number of PDEs
U=zeros(d,NI); % create correct sized array (without ghost cells)
dx=xcen(2)-xcen(1);
lengthofdomain=NI*dx;
xmidpoint=lengthofdomain/2;
g=9.81;
%
% 1D Dam break.  Left and right states are either side 
% of the centre of computational domain.
% Left states
hleft=10.0; % left water depth [m]
vleft=[0,0]; % left water velocity vector [m/s]
phileft=hleft*g; % left geopotential
% Right states
hright=5.0; % right water depth [m]
vright=[0,0]; % right water velocity vector [m/s]
phiright=hright*g; % right geopotential
%
% Fill in U values for each computational cell
for i=1:NI
    if (xcen(i)<=xmidpoint)
        U(1,i)=phileft;
        U(2,i)=phileft*vleft(1);
    else
        U(1,i)=phiright;
        U(2,i)=phiright*vright(1);
    end % of if
end % of i loop
%
end % finitialU
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%% 
function U=finsertghostcells(U)
% Verification:  verified.
% Inserts ghost cells at left and right ends of the computational domain. 
% Hence computational indices go from i=2 to NI+1.
% Assumes that U has 2 rows (i.e. 2 PDEs) which is correct for the 1D SWE.
% 
 [d,NI]=size(U); % d is the number of equations (= rows of U).
 leftghost=[0
            0];
%
 rightghost=[0
             0];
%
 U=[leftghost U rightghost];
end % finsertghostcells
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%% 
function U=fbcs(U)
% Verification:  verified.
% Inserts boundary conditions in to the ghost cells at left and right ends 
% of the domain.  Note that ghost cells have already been appended to U and
% U is initially zero in each ghost cell.
% 
% The following applies zero gradient boundary conditions everywhere.
 [d,N]=size(U); % d is the number of equations (= rows of U).
% Left
 U(:,1)=U(:,2); % copies the neighbouring interior values to the ghost cells
% Right
 U(:,N)=U(:,N-1); % copies the neighbouring interior values to the ghost cells
end % fbcs
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
function Unext=fsolver(U,A,dt,S,x)
% Implements the finite volume scheme for 1 time step
[d,N]=size(U); % includes ghost cells
NI=N-2; % number of computational cells
for i=2:NI+1 % loop over computational cells
     SRight=[S(i,1,1),S(i,1,2)]; % right side vector
     SLeft=[S(i,2,1),S(i,2,2)];
     [IHRight,IHLeft]=finterfaceflux(U,i,x); % finds the left and right 
     %                                       interface fluxes for cell i
     % Update U to Unext
     for k=1:d
     % compute total flux out of cell i for equation k
       IHR=[IHRight(k,1),IHRight(k,2)]; % right flux vector
       IHL=[IHLeft(k,1),IHLeft(k,2)];   % left flux vector
       totalfluxout=dot(IHR,SRight)+dot(IHL,SLeft);
       % Uses Lax-Friedrichs
       Unext(k,i)=0.5*(U(k,i-1)+U(k,i+1))-(dt/A(i))*totalfluxout; 
     end % k loop
 end % of i loop
 end % fsolver
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
function [IHRight,IHLeft]=finterfaceflux(U,i,x)
% Interface H values are based on linear interpolation of neighbouring cell 
% centre data which is the same as averaging on a uniform mesh.  
% These may be found either by finding the average value of U at the cell 
% at the interface and then constructing H or by direct averaging of H 
% components. Since the PDEs are non-linear these approaches are not the same.
% The latter appoach is analogous to the Lax-Friedrichs finite 
% difference method and is the better method as the former method produces
% unwanted oscillations.
% NEEDS GENERALISING TO NON CARTESIAN MESH USING INTERPOLATION
%
% Interface fluxes are denoted IHRight, IHLeft.
% For the 1D shallow water equations H is:
% [     phi*vx          i  +  [ 0     j
%    phi*vx^2+phi^2/2 ]         0 ]    

[d,m]=size(U);   % d=number of rows of U (= number of PDEs = 2)
NI=m-2;  % number of computational cells
IHRight=zeros(d,2); % array to store the right interface flux for a cell. 
                    % IHRight(,1) is the i component of H and
                    % IHRight(,2) is the j component of H which is zero.
IHLeft=zeros(d,2);
%
% Find interface values of H directly by averaging neighbouring components.
%
% Cell i
Hi=fgetHifromUi(U,i);
% Cell i-1
Himinus1=fgetHifromUi(U,i-1);
% Cell i plus 1
Hiplus1=fgetHifromUi(U,i+1);
%
% Right interface H components are obtained by linear extrapolation of 
% components of H at the centre of cell i.
%
dx=x(2,1)-x(1,1); % length of uniform Cartesian cell in i (x) direction.
d1=dx/2; % Distance from centre of cell i to interface i+1/2
         % measured along the line joining the centres of cells i and i+1.
d2=dx;   % Distance between centres of cells i and i+1.
IGradH=(Hiplus1-Hi)/d2; % gradients of H components in i direction
IHRight=Hi + d1*IGradH;
%
% Left interface H components are obtained by linear extrapolation of 
% components of H at the centre of cell i-1.
%
d1=dx/2; % Distance from centre of cell i-1 to interface i-1/2
         % measured along the line joining the centres of cells i-1 and i.
d2=dx;   % Distance between centres of cells i-1 and i.
IGradH=(Hi-Himinus1)/d2; % gradients of H components in i direction
IHLeft=Himinus1 + d1*IGradH;
%
end % finterfaceflux
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%% 
function outarray=fremoveghostcells(inarray)
% Verification:  Verified
% Removes left and right ghost cells.
 [d,m]=size(inarray);
 outarray=inarray(:,2:m-1);
end % fremoveghostcells
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
function dt=fcalcdt(A,S,U)
% Verification:  not verified.
% Finds allowable time step dt using heuristic formula.
% 0<F<1 is a tunable safety factor.
 [nrows,M]=size(U);  % U now enlarged to include ghost cells 
 dt=9999999;  % large value of dt to start
 F=0.9; % safety factor
 for i=2:M-1 % indices go over computational cells 
     phi=U(1,i);
     cel=sqrt(phi); % wave celerity
     vx=U(2,i)/phi; % flow speed in x direction
     v=[vx,0]; % flow velocity
     SRight=[S(i,1,1),S(i,1,2)];
     a=dot(v,SRight);
     b=cel*norm(SRight);
     %
     speed=max(abs(a+b),abs(a-b));  % max abs wave speed
     localdt=A(i)/speed;
     % take smallest value of dt
     if (localdt<dt)
        dt=localdt;
     end
  end % i loop
 dt=F*dt;
end  % fcalcdt
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
function fplotresults(xcen,U)
% Plots results on a structured mesh - even works for non-Cartesian!
%
  g=9.81;  % acceleration due to gravity
  waterdepth=U(1,:)/g;
  waterspeed=U(2,:)./U(1,:);
  subplot(2,1,1), plot(xcen,waterdepth)
  title('FVSWE1Dv3.m: numerical solution for water depth')
  xlabel('x [m]')
  ylabel('water depth [m]')
  %
  subplot(2,1,2), plot(xcen,waterspeed)
  title('FVSWE1Dv3.m: numerical solution for x component of water velocity')
  xlabel('x [m]')
  ylabel('speed [m/s]')
%
end % fplotresults
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
function H=fgetHifromUi(U,i)
% Constructs flux vector H at centre of cell i from U(i) 
% U = [  phi
%       fi*vx ]
%
% H=[     phi*vx    ,  0          Each row of H is a 2D vector 
%    phi*vx^2+fi^2/2,  0 ]
%
[d,n]=size(U); % d is the number of PDEs (= rows of U)
H=zeros(d,2);  % flux density vectors
%
phi=U(1,i); % geopotential
vx=U(2,i)/phi; % velocity component in x direction
%
H(1,1)=U(2,i); % i component of first row of H
H(1,2)=0;      % j component of first row of H
%
H(2,1)=phi*vx*vx+0.5*phi*phi; % i component of second row of H
H(2,2)=0;                     % j component of second row of H
%
end % fgetHifromUi
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%







% %function FVSWELAXF2D
% % File name: FVSWELAXF2D.m
% %
% % Compute mesh data (includes ghost cells):
% [x,y,A,xcen,ycen,solid,R,LI,Iu,LJ,Ju,LIl,LIr,LJl,LJr]=freadmesh(Ndim);
% disp('mesh read in and mesh data computed')
% %
% S=fsidevectors(x,y); % compute and store cell side vectors in an
%                      % (NI+2)x(NJ+2)x4x2 array which contains ghost cells.
% 
% %%
% disp('time marching starts')   
% for timestep=1:maxtimesteps 
%  U=fbcs(U);  % Implement boundary conditions using ghost cells. 
%              % U is (NI+2)x(NJ+2)x3. Computational cells are 
%              % indexed i=2 to NI+1, j=2 to NJ+1.
%  [iGrad,jGrad,iGradL,iGradR,jGradL,jGradR]=fgradients(U,LI,Iu,LJ,Ju,LIl,LIr,LJl,LJr); % gets gradient vectors
%  IH=fIflux(scheme,U,iGrad,jGrad,iGradL,iGradR,jGradL,jGradR,Iu,Ju,R,S); % gets all interface flux vectors
%  dt=fcalcdt(Ndim,A,S,U); % Finds stable time step for each iteration.
%  t=t+dt      % output current time 
%  if (t>runtime)
%      dt=runtime-(t-dt); % reduce last time step to stop at t = runtime
%      t=runtime;
%      stop=1; % flag to stop program at t = runtime
%  end
%  timestep  % output current time step
%  for j=2:NJ+1 
%    for i=2:NI+1 
%        %
%        s1=[S(i,j,1,1),S(i,j,1,2)]; % side vector for side 1
%        s2=[S(i,j,2,1),S(i,j,2,2)]; % side vector for side 2
%        s3=[S(i,j,3,1),S(i,j,3,2)]; % side vector for side 3
%        s4=[S(i,j,4,1),S(i,j,4,2)]; % side vector for side 4
%        %
%        dtoverA=dt/A(i,j);
%        %
%      for k=1:d
%        %IH is (NI+2)x(NJ+2)xdx4x2;
%        IH1=[IH(i,j,d,1,1),IH(i,j,d,1,2)]; % interface flux vector for side 1
%        IH2=[IH(i,j,d,2,1),IH(i,j,d,2,2)]; % interface flux vector for side 2
%        IH3=[IH(i,j,d,3,1),IH(i,j,d,3,2)]; % interface flux vector for side 3
%        IH4=[IH(i,j,d,4,1),IH(i,j,d,4,2)]; % interface flux vector for side 4      
%        % FV scheme
%        % compute total flux out of cell (i,j) for row k
%        totalfluxout=dot(IH1,s1)+dot(IH2,s2)+dot(IH3,s3)+dot(IH4,s4);
%        Unext(i,j,k)=U(i,j,k)-dtoverA*totalfluxout; 
%      end % k loop
% %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% function [x,y,A,xcen,ycen,solid,R,LI,Iu,LJ,Ju,LIl,LIr,LJl,LJr]=freadmesh(Ndim)
% % The mesh is structured and has NI computational cells in the i direction 
% % and NJ computational cells in the j direction (so each cell has 4 sides).
% % The x and y coordinates of the lower left hand corner of cell (i,j) are
% % held in arrays x and y respectively which are both (NI+1)x(NJ+1) before 
% % ghost cells re appended.  In this way the 4 vertices of cell (i,j) are 
% % (x(i),y(j)), (x(i+1),y(j)), (x(i+1),y(j+1)) and (x(i),y(j+1)). A holds cell
% % areas, xcen and ycen hold cell centres and solid holds solid flags - if cell (i,j) 
% % is solid then solid(i,j)=1 otherwise solid(i,j)=0.
% % R holds normal vectors from cell centres to each of the 4 cell faces.
% % Ghost cells are appended to all arrays increasing each dimension index by 2.
% %
% % NI=51,NJ=51,dx=2,dy=2 makes mesh centre (51,51)
% NI=10;
% NJ=51;
% dx=2;
% dy=2;
% if (Ndim==1)
%     % 1D problem
%     NJ=1;
%     dy=1;
% end
% x=zeros(NI+1,NJ+1); % allocate correct sized array
% y=zeros(NI+1,NJ+1); % allocate correct sized array
% %
% % Create ghost cells around whole mesh (for gradient calcs)
% % First embed x and y in larger arrays
% tempx=zeros(NI+3,NJ+3);
% tempx(2:NI+2,2:NJ+2)=x;
% x=tempx;
% clear tempx
% tempy=zeros(NI+3,NJ+3);
% tempy(2:NI+2,2:NJ+2)=y;
% y=tempy;
% clear tempy
% % Now fill in ghost cell vertices by reflecting cell vertices
% % i.e. if z is the vector from vertex (i,2) to vertex (i,3) then 
% % ghost vertex (i,1) has position vector (i,2)-z = 2(i,2)-(i,3)
% for i=2:NI+2
%     x(i,1)=2*x(i,2)-x(i,3);
%     y(i,1)=2*y(i,2)-y(i,3);
%     x(i,NJ+3)=2*x(i,NJ+2)-x(i,NJ+1);
%     y(i,NJ+3)=2*y(i,NJ+2)-y(i,NJ+1);
% end
% %
% for j=2:NJ+2
%     x(1,j)=2*x(2,j)-x(3,j);
%     y(1,j)=2*y(2,j)-y(3,j);
%     x(NI+3,j)=2*x(NI+2,j)-x(NI+1,j);
%     y(NI+3,j)=2*y(NI+2,j)-y(NI+1,j);
% end
% %
% % Find cell areas
% A=zeros(NI+2,NJ+2);  % allocate correct size array
% for i=1:NI+2
%     for j=1:NJ+2
%         d1=[x(i+1,j+1)-x(i,j),y(i+1,j+1)-y(i,j),0];
%         d2=[x(i,j+1)-x(i+1,j),y(i,j+1)-y(i+1,j),0];
%         d=cross(d1,d2);   
%         A(i,j)=0.5*sqrt(dot(d,d));
%     end % of j loop
% end % of i loop
% %
% % Find cell centres by averaging coordinates of vertices
% xcen=zeros(NI+2,NJ+2);  % allocate correct sized array
% ycen=zeros(NI+2,NJ+2);  % allocate correct sized array
% for i=1:NI+2
%     for j=1:NJ+2
%         xcen(i,j)=(x(i,j)+x(i+1,j)+x(i+1,j+1)+x(i,j+1))/4;
%         ycen(i,j)=(y(i,j)+y(i+1,j)+y(i+1,j+1)+y(i,j+1))/4;
%     end % of j loop
% end % of i loop
% %
% % Find normal vector from cell centre to each of 4 cell faces
% % using a vector based formula R=p-q, where q is the cell centre position
% % vector, p is the position vector of the point on the cell side,
% % p = a + t (b-a) where a and b are position vectors of cell vertices
% % defining the side and t = (q-a).(b-a)/(b-a).(b-a)
% R=zeros(NI+2,NJ+2,4,2);  % create correct sized array
% for i=1:NI+2
%     for j=1:NJ+2
%         q=[xcen(i,j),ycen(i,j)];
%         % side 1
%         a=[x(i,j),y(i,j)];
%         b=[x(i+1,j),y(i+1,j)];
%         bminusa=b-a;
%         qminusa=q-a;
%         t=dot(qminusa,bminusa)/dot(bminusa,bminusa);
%         p=a+t*bminusa;
%         answer=p-q;
%         R(i,j,1,1)=answer(1);
%         R(i,j,1,2)=answer(2);
%         % side 2
%         a=[x(i+1,j),y(i+1,j)];
%         b=[x(i+1,j+1),y(i+1,j+1)];
%         bminusa=b-a;
%         qminusa=q-a;
%         t=dot(qminusa,bminusa)/dot(bminusa,bminusa);
%         p=a+t*bminusa;
%         answer=p-q;
%         R(i,j,2,1)=answer(1);
%         R(i,j,2,2)=answer(2);
%         % side 3
%         a=[x(i+1,j+1),y(i+1,j+1)];
%         b=[x(i,j+1),y(i,j+1)];
%         bminusa=b-a;
%         qminusa=q-a;
%         t=dot(qminusa,bminusa)/dot(bminusa,bminusa);
%         p=a+t*bminusa;
%         answer=p-q;
%         R(i,j,3,1)=answer(1);
%         R(i,j,3,2)=answer(2);
%         % side 4
%         a=[x(i,j+1),y(i,j+1)];
%         b=[x(i,j),y(i,j)];
%         bminusa=b-a;
%         qminusa=q-a;
%         t=dot(qminusa,bminusa)/dot(bminusa,bminusa);
%         p=a+t*bminusa;
%         answer=p-q;
%         R(i,j,4,1)=answer(1);
%         R(i,j,4,2)=answer(2);
%     end % of j loop
% end % of i loop
% % 
% % Find:
% % LI is the distance from centre of cell (i-1,j) to centre of 
% % cell (i+1,j). Iu(i,j) holds the unit vector. Similarly for LJ and Ju. 
% % These are used to give a limited gradient vectors.
% LI=zeros(NI+2,NJ+2); % create correct sized array
% Iu=zeros(NI+2,NJ+2,2); % create correct sized array
% LJ=zeros(NI+2,NJ+2); % create correct sized array
% Ju=zeros(NI+2,NJ+2,2); % create correct sized array
% for i=2:NI+1 % indices for computational cells
%     for j=2:NJ+1 % indices for computational cells
%         % i direction
%         a=[xcen(i-1,j),ycen(i-1,j)];
%         b=[xcen(i+1,j),ycen(i+1,j)];
%         n=b-a;  % vector joining cell centres
%         LI(i,j)=sqrt(dot(n,n)); % length of vector n
%         nunit=n/LI(i,j); % unit vector joining cell centres
%         Iunit(i,j,1)=nunit(1); % store i component
%         Iunit(i,j,2)=nunit(2); % store j component
%         % j direction
%         a=[xcen(i,j-1),ycen(i,j-1)];
%         b=[xcen(i,j+1),ycen(i,j+1)];
%         n=b-a;  % vector joining cell centres
%         LJ(i,j)=sqrt(dot(n,n)); % length of vector n
%         nunit=n/LJ(i,j); % unit vector joining cell centres
%         Junit(i,j,1)=nunit(1); % store i component
%         Junit(i,j,2)=nunit(2); % store j component       
%     end % of j loop
% end % of i loop
% %
% % Find:
% % LIl is the distance between the centres of cells (i-1,j) and (i,j)
% % LIr is the distance between the centres of cells (i,j) and (i+1,j)
% % these are used for finding left and right I gradients in cell (i,j)
% % LJl is the distance between the centres of cells (i,j-1) and (i,j)
% % LJr is the distance between the centres of cells (i,j) and (i,j+1)
% % these are used for finding left and right J gradients in cell (i,j)
% LIl=zeros(NI+2,NJ+2); % create correct sized array
% LIr=zeros(NI+2,NJ+2); % create correct sized array
% LJl=zeros(NI+2,NJ+2); % create correct sized array
% LJr=zeros(NI+2,NJ+2); % create correct sized array
% for i=2:NI+1 % indices for computational cells
%     for j=2:NJ+1 % indices for computational cells
%         % i direction
%         a=[xcen(i-1,j),ycen(i-1,j)];
%         b=[xcen(i,j),ycen(i,j)];
%         c=[xcen(i+1,j),ycen(i+1,j)]; 
%         leftvector=b-a;
%         rightvector=c-b;
%         LIl(i,j)=sqrt(dot(leftvector,leftvector));
%         LIr(i,j)=sqrt(dot(rightvector,rightvector));
%         % j direction
%         a=[xcen(i,j-1),ycen(i,j-1)];
%         b=[xcen(i,j),ycen(i,j)];
%         c=[xcen(i,j+1),ycen(i,j+1)]; 
%         lowervector=b-a;
%         uppervector=c-b;
%         LJl(i,j)=sqrt(dot(lowervector,lowervector));
%         LJr(i,j)=sqrt(dot(uppervector,uppervector));
%     end % of j loop
% end % of i loop
% %       
% end % freadmesh
% %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% function IH=fIflux(scheme,U,iGrad,jGrad,iGradL,iGradR,jGradL,jGradR,Iu,Ju,R,S)
% % Calculates the 4 interface fluxes on the 4 sides of cell (i,j) 
% % for each component of U.
% % Cell sides are numbered 1 to 4 anti-clockwise.  For cell (i,j) side 1 
% % is the vector s1 from (x(i,j), y(i,j)) and to (x(i+1,j),y(i+1,j)).
% % This means that sides 2 and 4 are in the i direction and sides 1 and 3
% % are in the j direction.
% % iGrad and jGrad are gradient vectors in i and j directions.
% % R holds normal vectors from cell centres to each of the 4 cell faces.
% % S holds side vectors for each of the 4 cell faces.
% % Interface fluxes are denoted IH1, IH2, IH3 and IH4 and are 2D vectors.

% % 
% [m,n,d]=size(U);  % d=number of rows of U
% NI=m-2; NJ=n-2;  % number of computational cells in i and j directions
% IH=zeros(NI+2,NJ+2,d,4,2); % array to store the 4 interface flux vectors 
% %                            for each cell.
% %
% % 

%   H=getHfromU(U); % gets H at all cell centres from U values there

% % case 'HighRes'
% %   %% HighRes scheme
% %   disp('HighRes scheme')
% %     for i=2:NI+1
% %      for j=2:NJ+1      
% %         % cell side 2: interface (i+1/2,j) 
% %           s=[S(i,j,2,1),S(i,j,2,2)];
% %         % Left side of interface (i+1/2,j)      
% %           r=[R(i,j,2,1),R(i,j,2,2)];
% %           iG=[iGrad(i,j,1),iGrad(i,j,2)]; % u gradient vector in i direction
% %           uL=u(i,j)+dot(r,iG); % piecewise linear reconstruction
% %         % Right side of interface (i+1/2,j)              
% %           r=[R(i+1,j,4,1),R(i+1,j,4,2)];
% %           iG=[iGrad(i+1,j,1),iGrad(i+1,j,2)]; % u gradient vector in i direction
% %           uR=u(i+1,j)+dot(r,iG); % piecewise linear reconstruction
% %           %
% %           Hstar=fRiemann(uL,uR,v,s); % Riemann interface flux
% %           IH(i,j,2,1)=Hstar(1);  % component of IH2 in the x direction
% %           IH(i,j,2,2)=Hstar(2);  % component of IH2 in the y direction
% %         %  
% %         % cell side 4: interface (i-1/2,j) 
% %           s=-[S(i,j,4,1),S(i,j,4,2)]; % reverse direction 
% %         % Left side of interface (i-1/2,j)         
% %           r=[R(i-1,j,2,1),R(i-1,j,2,2)];
% %           iG=[iGrad(i-1,j,1),iGrad(i-1,j,2)]; % u gradient vector in i direction
% %           uL=u(i-1,j)+dot(r,iG); % piecewise linear reconstruction          
% %         % Right side of interface (i-1/2,j)
% %           r=[R(i,j,4,1),R(i,j,4,2)];
% %           iG=[iGrad(i,j,1),iGrad(i,j,2)]; % u gradient vector in i direction
% %           uR=u(i,j)+dot(r,iG); % piecewise linear reconstruction
% %           %
% %           Hstar=fRiemann(uL,uR,v,s); % Riemann interface flux
% %           IH(i,j,4,1)=Hstar(1);  % component of IH4 in the x direction
% %           IH(i,j,4,2)=Hstar(2);  % component of IH4 in the y direction
% %         %
% %         % cell side 3: interface (i,j+1/2)
% %           s=[S(i,j,3,1),S(i,j,3,2)];
% %           % Left side of interface (i,j+1/2)
% %            r=[R(i,j,3,1),R(i,j,3,2)];
% %            jG=[jGrad(i,j,1),jGrad(i,j,2)]; % u gradient vector in j direction
% %            uL=u(i,j)+dot(r,jG); % piecewise linear reconstruction 
% %           % Right side of interface (i,j+1/2)  
% %            r=[R(i,j+1,1,1),R(i,j+1,1,2)];
% %            jG=[jGrad(i,j+1,1),jGrad(i,j+1,2)]; % u gradient vector in j direction
% %            uR=u(i,j+1)+dot(r,jG); % piecewise linear reconstruction
% %            %
% %            Hstar=fRiemann(uL,uR,v,s); % Riemann interface flux
% %            IH(i,j,3,1)=Hstar(1);  % component of IH3 in the x direction
% %            IH(i,j,3,2)=Hstar(2);  % component of IH3 in the x direction
% %         %
% %         % cell side 1
% %           s=-[S(i,j,1,1),S(i,j,1,2)]; % reverse direction
% %           % Left side of interface (i,j-1/2)
% %            r=[R(i,j-1,3,1),R(i,j-1,3,2)];
% %            jG=[jGrad(i,j-1,1),jGrad(i,j-1,2)]; % u gradient vector in j direction
% %            uL=u(i,j-1)+dot(r,jG); % piecewise linear reconstruction
% %            % Right side of interface (i,j-1/2)
% %            r=[R(i,j,1,1),R(i,j,1,2)];
% %            jG=[jGrad(i,j,1),jGrad(i,j,2)]; % u gradient vector in j direction
% %            uR=u(i,j)+dot(r,jG); % piecewise linear reconstruction
% %            %
% %            Hstar=fRiemann(uL,uR,v,s); % Riemann interface flux
% %            IH(i,j,1,1)=Hstar(1);  % component of IH1 in the x direction
% %            IH(i,j,1,2)=Hstar(2);  % component of IH1 in the y direction
% %            %
% %      end % of j loop
% %     end % of i loop  
% %   %% end of HighRes scheme
% %  case 'Halfhancock'
% %      disp('Halfhancock scheme')
% %     % Method is based on HighRes and includes limiters on gradients.
% %     % First find limited gradient vectors iG, jG in each cell.
% %     for i=2:NI+1
% %      for j=2:NJ+1
% %          % I direction
% %          g=flimiter(iGradR(i,j),iGradL(i,j)); % limited u gradient
% %          iG=g*[Iu(i,j,1),Iu(i,j,2)]; % u gradient vector in i direction
% %          % J direction
% %          g=flimiter(jGradR(i,j),jGradL(i,j)); % limited u gradient
% %          jG=g*[Ju(i,j,1),Ju(i,j,2)]; % u gradient vector in j direction
% %      end
% %     end
% %     % Now find left and right u values at interfaces and solve 
% %     % the associated Riemann problem.
% %     for i=2:NI+1
% %      for j=2:NJ+1      
% %         % cell side 2: interface (i+1/2,j) 
% %           s=[S(i,j,2,1),S(i,j,2,2)];
% %         % Left side of interface (i+1/2,j)      
% %           r=[R(i,j,2,1),R(i,j,2,2)];
% %           uL=u(i,j)+dot(r,iG); % piecewise linear reconstruction
% %         % Right side of interface (i+1/2,j)              
% %           r=[R(i+1,j,4,1),R(i+1,j,4,2)];
% %           uR=u(i+1,j)+dot(r,iG); % piecewise linear reconstruction
% %           %
% %           
% %     end % of j loop
% %    end % of i loop    

% end % fIflux 
% %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%% 
% function [iGrad,jGrad,iGradL,iGradR,jGradL,jGradR]=fgradients(U,LI,Iu,LJ,Ju,LIl,LIr,LJl,LJr)
% % Verification: 
% %
% % Finds the gradients for each component of U in the i and j directions for each cell.
% % iGrad(i,j,d) = [(U(i+1,j,d)-U(i-1,j,d))/LI(i,j)] Iu(i,j,:).
% % iGradL(i,j,d) = [(U(i,j,d)-U(i-1,j,d))]/LIl(i,j).  etc.
% % (ghost cell gradients are zero). 
% %
% [m,n,d]=size(U); % d is number of PDEs
% NI=m-2; NJ=n-2;  % number of computational cells in i and j directions
% iGrad=zeros(NI+2,NJ+2,3,2); % gradient vector
% jGrad=zeros(NI+2,NJ+2,3,2); % gradient vector
% iGradL=zeros(NI+2,NJ+2,3); % scalar gradient
% jGradL=zeros(NI+2,NJ+2,3); % scalar gradient
% iGradR=zeros(NI+2,NJ+2,3); % scalar gradient
% jGradR=zeros(NI+2,NJ+2,3); % scalar gradient
% Udiff=zeros(d);
% Udiffoverd=zeros(d);
% for i=2:NI+1
%     for j=2:NJ+1
%         % i direction 
%         % gradient vectors
%       for k=1:d
%         Udiff(k)=U(i+1,j,k)-U(i-1,j,k);
%         Udiffoverd(k)=Udiff(k)/LI(i,j);
%         iGrad(i,j,k,1)=Udiffoverd(k)*Iu(i,j,1);
%         iGrad(i,j,k,2)=Udiffoverd(k)*Iu(i,j,2);
%         % left and right scalar gradients
%         iGradL(i,j,k)=(U(i,j,k)-U(i-1,j,k))/LIl(i,j);
%         iGradR(i,j,k)=(U(i+1,j,k)-U(i,j,k))/LIr(i,j);
%         % j direction
%         % gradient vectors
%         Udiff(k)=U(i,j+1,k)-U(i,j-1,k);
%         Udiffoverd(k)=Udiff(k)/LJ(i,j);
%         jGrad(i,j,k,1)=Udiffoverd(k)*Ju(i,j,1);
%         jGrad(i,j,k,2)=Udiffoverd(k)*Ju(i,j,2);
%         % lower and upper scalar gradients
%         jGradL(i,j,k)=(U(i,j,k)-U(i,j-1,k))/LJl(i,j);
%         jGradR(i,j,k)=(U(i,j+1,k)-U(i,j,k))/LJr(i,j); 
%       end % for loop
%     end % of j loop
% end % of i loop
% %
% end % fgradients
